PAGE    120,160
title   (DE.ASM - DOS Command Line Editor)
;***************************************************************************
;       The first time this program is executed, it makes itself memory 
;       resident.  If this program is subsequently executed with no 
;       parameters, and no other TSR has grabbed int 21h, it does nothing.
;       If it is subsequently executed with /R as the parameter, it
;       deinstalls the first memory resident DE.
;
;***************************************************************************
;   This code has been modified and enhanced by Tim Riker. (TJR)
;   this is ver 0.02.
;   the following DEFINES were added:
;***************************************************************************
;   LOGFILE: makes a /f:filepath on command line log all commands with
;       date and time to a log file.  (primative but functional)
;   FNKEY: defines function keys.
;   NOTSOME: excludes certian commands from being buffered:
;       exit, dir/<x>, cls are currently coded.
;   AUX: allows buffering of non-command line input (ex: dos DEBUG commands)
;   AUXBUF: sets up separate buffer for non-command line input
;       TJR: now uses length of buffer to determine source as DOS 4.01 seems
;           to put buffer in high memory instead of below our CS.
;   LINECAP: PgUp/PgDn do screen line capture.  This code was taken from
;       NEWDOSED.COM also by Jack Gersbeck.
;       TJR: now supports 43/50/etc line mode through BIOS table
;   NDEBUG: defined to remove debugging code
;   NOTUSED: just to comment out stuff that seems unused
;
;***************************************************************************
; General comments   (as in original source from Scott W. Killen
;***************************************************************************
; The following comprises the disassembled and documented version of DOS-ED
; as it exists on Compuserve's XA6 data base.  If one reassembles this
; program one will obtain the exact version of DOS-ED.COM that exists on
; that database.  No attempt has been made to make improvements to this code
; other than to indicate in the margin were some could (and probably should)
; be made.  If you notice that the program sizes are different between this
; reassembled and relinked version and the version that has been generated
; from the HEX file on CIS, that is because there is some trailing garbage
; on the CIS file.  You can strip it off in the debugger and the files will
; compare exactly.
;
; Even after disassembling and documenting this code I have a few questions
; about what was being done in places.  If my comments seem spurious at
; times, then just take them with a grain of salt.  They should help you
; understand this code for the most part.
;
; When linking this program you should expect to receive a No Stack warning
; message.  This is normal and should be ignored.
;
; DOSEDIT was written by Jack Gersbach.
; This disassembly and documentation was done by myself.
;                                     - Scott W. Killen
;
;***************************************************************************
; Definitions and commentary on conventions in coding
;***************************************************************************
; Register usage:  (Most common (but not exclusive) uses are listed)
; AX, DX, CX - Used for many reasons
; DH - Usually current row on crt
; DL - Usually current col on crt
; SI - Usually points to offset of current character in input buffer 
;       (first character is offset 0)
; DI - Usually has the value of one of the following:
;      * The number of characters currently in the input buffer   *or*
;      * Points to a destination in the text holding buffer
; BX - Usually contains the address of the first character in the input buffer.
;         (The two characters before this are the max buffer size and length 
;       of string characters.)
; BP - Usually a register of flags organized as bits |01234567| where:
;      Bits 0-2 have no function
;      Bit 3 - Off means no line marker on screen
;              On means line marker on screen
;      Bit 4 - Off means the input buffer has been altered since interrupt 
;               processing began.
;              On means *only* lateral cursor positioning commands have been 
;               issued since interrupt processing began.
;      Bit 5 - Off means previous line from holding buffer, On means original 
;               line
;      Bit 6 - Off means no changes to current line in input buffer, On means 
;               changes have been made
;      Bit 7 - Applies to the redisplay of lines on the screen after an 
;               intra-line edit operation
;                 Off means line wasn't shortened therefore no trailing 
;                       blanks needed
;                 On means line is now shorter and blanks are needed over 
;                       trailing characters
;
; Memory Usage:
; ES:[0450H] - Contains current cursor position in Bios data area
; ES:[044AH] - Contains # of columns of current crt screen in Bios data area
; ES:[0484H] - Contains # of rows of current crt screen in Bios data area
; ES:[0417H] - Keyboard flag, containing insert, shift, num lock, scroll lock 
;               etc in Bios data area
;                   insert flag bit is 80h
;
; CS:[inline_mod1+1] - Jump return address to normal DOS funcion processing
; CS:[inline_mod2+1] - Current cursor definition (start/stop lines)
; CS:[inline_mod3+1] - Screen location of end of line (row/col)
; CS:[inline_mod4+2] - Higher of DS or CS on entry into interrupt.
; CS:[inline_mod5+2] - Maximum size of input buffer
;
; start_off  - Offset of start of input buffer on screen (example if prompt is
;               A> then screen column position of input buffer start would be 2
; current_col - Current column position on screen (1st column is 0)
; last_entry - Address (offset) of last string accessed in text buffer
; next_entry - Address (offset) of next available block in text buffer
; buffer_bot - Base address of text buffer
; buffer_top - Top address of text buffer
; aux_last  \
; aux_next   \ Same as previous four except for auxiliary buffer
; aux_bot    /
; aux_top   /
;
;***************************************************************************
dd_template     struc                           ; to access a double word
l               dw      ?
h               dw      ?
dd_template     ends

IFDEF AUXBUF
IFNDEF AUX
.ERR
%OUT AUX must be defined to use AUXBUF
ENDIF
ENDIF
;       Character equates
ACK                 equ 06                      ; Acknowledge character (Cntrl F)
BELL                equ 07                      ; Bell character        (Cntrl G)
BACKSPACE           equ 08                      ; Backspace character
TAB                 equ 09                      ; Tab character
ESCAPE              equ 1Bh                     ; Escape character
BLANK               equ 20h                     ; Space character
CR                  equ 0Dh                     ; carriage return
LF                  equ 0Ah                     ; Line feed
HAT                 equ 5Eh                     ; "^" character
SMALLA              equ 61h                     ; "a" character
LEFT_CURLY_BRACE    equ 7Bh                     ; "{" character
CNTL_BACKSPACE      equ 7Fh                     ; Control & backspace pressed simultaneously

;       Masks
INSERT_ON_MASK      equ 80h                     ; Insert key bit on, corresponding to keyboard
                                                ; status word
INSERT_OFF_MASK     equ 7fh                     ; Insert key bit off, corresponding to keyboard
                                                ; status word
LENGTH_BYTE_MASK    equ 80h                     ; Mask to identify the text string length bit
                                                ; in text buffer
LENGTH_OFF_MASK     equ 7fh                     ; Mask to turn of the length byte identifier
UPPER_CASE_MASK     equ 40h                     ; Convert control char to upper case char
LOWER_CASE_MASK     equ 20h                     ; Convert upper case char to lower case char

;       Physical characteristics
ROWS_ON_CRT         equ 19h                     ; Ordinal number of lines on crt
LAST_ROW_ON_CRT     equ ROWS_ON_CRT - 1         ; Offset of last line on crt
                                                ; (top line is line 0)
MONO_CUR_DEF        equ 0B0Ch                   ; Start, stop lines of cursor
MONO_CUR_STOP       equ 0Ch                     ; Stop line for standard cursor
GRAPHICS_CUR_DEF    equ 0607h                   ; Start, stop lines of cursor for graphics device
GRAPHICS_CUR_STOP   equ 07h                     ; Stop line for standard cursor for graphics device

DOS_BUF_CHARS       equ 7fh                     ; Number of characters in a COMMAND.COM request
                                                ; Note BYTE PTR DS:DX = 80h which includes the CR

; Bios interrupt 10 functions
SET_CURSOR_TYPE     equ 01h
SET_CURSOR_POSN     equ 02h
READ_CURSOR_POSN    equ 03h
READ_VIDEO_STATE    equ 0fh
;
; Dos interrupt 21 functions
INPUT_CHR_NOECHO    equ 08h
OUTPUT_CHR          equ 02h
GET_INT_VECTOR      equ 35h
SET_INT_VECTOR      equ 25h
GET_PSP_ADDR        equ 62h
SET_PSP_ADDR        equ 50h
DOS_OPEN_H          equ 3dh                     ; open file (handle mode)
DOS_CREATE_NEW_H    equ 5bh
DOS_WRITE_H         equ 40h
DOS_CLOSE_H         equ 3eh
DOS_SET_FILE_PTR    equ 42h
DOS_GET_DATE        equ 2ah
DOS_GET_TIME        equ 2ch
;
;***************************************************************************
; BIOS data segment definition
;***************************************************************************
bioseg          segment at 0h
                org     417h
kbd_flag        db      ?
                org     44ah
crt_cols        dw      ?
                org     450h
cursor_posn     dw      ?
                org     484h
crt_rows        db      ?                       ; last row on screen 18h for 25 line modes
bioseg          ends
;
;***************************************************************************
; DOSEDIT code segment begins
;***************************************************************************
cseg            segment 'CODE'
                assume cs:cseg
;
                ORG 100H
dosedit         PROC NEAR
                JMP     install                 ; Now go do it.

                org     16h
parent_psp      dw      ?
file_table      db      20h dup (?)
                org     2Ch
envrnmt_seg     dw      ?                       ; segment of program's environment segment

                org     34h
file_tbl_vec    dd      ?                       ; pointer to file handle table
                                                ; (normally at cs:0018h)
IFNDEF LOGFILE
                org     80h
paramline       db      80h dup (?)             ; address in PSP of command line parameters
ENDIF

                org     5Ch
;fcb1           label   word                    ; offset in PSP of first unopened FCB
                                                ; safe to start our buffer here if we don't
                                                ; use FCB calls!
deflag          dw      ?                       ; holds "DE" which stands for DosEdit
oldint21vec     dd      ?                       ; holds old int 21 vector

IFDEF LINECAP
cur_cursor_pos  dw      ?                       ; high, low = row, col
ENDIF

IFDEF FNKEY
fk_ptr          dw      ?                       ; pointer used for function key routines
ENDIF

IFDEF LOGFILE
cur_psp         dw      ?                       ; PSP of calling program
file_name_ptr   dw      ?

                org     80h                     ; save the parameters!
paramline       db      80h dup (?)             ; address in PSP of command line parameters
ENDIF

puffer          label   byte                    ; start of buffer to hold commands
;***************************************************************************
BUF_SIZE         equ 100h

TEXT_BUF_START   equ offset puffer              ; Start of text holding buffer
                                                ; Size of text holding buffer in BUF_SIZE
TEXT_BUF_END     equ TEXT_BUF_START + BUF_SIZE  ; End of text buffer

IFDEF AUX
  IFDEF AUXBUF
    AUX_BUF_START  equ TEXT_BUF_START + BUF_SIZE; Start of auxiliary holding buffer
                                                ; Size of auxiliary holding buffer in BUF_SIZE
    AUX_BUF_END    equ AUX_BUF_START + BUF_SIZE ; End of auxiliary buffer
    ALL_BUF_SIZE   equ BUF_SIZE * 2             ; Combined buffer size
    ALL_BUF_END    equ TEXT_BUF_START + BUF_SIZE + BUF_SIZE
  ELSE
    ALL_BUF_SIZE   equ BUF_SIZE                 ; Only use text buffer
    ALL_BUF_END    equ TEXT_BUF_START + BUF_SIZE
  ENDIF
ELSE
  ALL_BUF_SIZE   equ BUF_SIZE                   ; Only use text buffer
  ALL_BUF_END    equ TEXT_BUF_START + BUF_SIZE
ENDIF


;***************************************************************************
;       Data area
;***************************************************************************
                org     ALL_BUF_END
start_off       db      00
current_col     db      00

last_entry      dw      TEXT_BUF_START
next_entry      dw      TEXT_BUF_START
buffer_bot      dw      TEXT_BUF_START
buffer_top      dw      TEXT_BUF_END

IFDEF AUXBUF
aux_last        dw      AUX_BUF_START
aux_next        dw      AUX_BUF_START
aux_bot         dw      AUX_BUF_START
aux_top         dw      AUX_BUF_END
ENDIF

IFDEF LOGFILE
line_feed       db      LF                      ; write on end of each log-line
ENDIF

;***************************************************************************
; Extended keystroke recognition and branching tables
;***************************************************************************
;  Table 1 of extended keyboard function codes
table1:
                db      0fh                     ; Back Tab
                db      47h                     ; Home key
                db      48h                     ; Up arrow
                db      4bh                     ; Left arrow
                db      4dh                     ; Right arrow
                db      4fh                     ; End key
                db      50h                     ; Down Arrow
                db      53h                     ; Del key
                db      73h                     ; Cntrl Left Arrow
                db      74h                     ; Cntrl Right Arrow
                db      75h                     ; Cntrl End
                db      77h                     ; Cntrl Home
                db      76h                     ; Cntrl PgDn
                db      84h                     ; Cntrl PgUp
IFDEF LINECAP
                db      49h                     ; PgUp
                db      51h                     ; PgDn
ENDIF

IFDEF FNKEY
                db      3Bh                     ; F1 key
                db      3Ch                     ; F2 key
                db      3Dh                     ; F3 key
                db      3Eh                     ; F4 key
                db      3Fh                     ; F5 key
                db      40h                     ; F6 key
                db      41h                     ; F7 key
                db      42h                     ; F8 key
                db      43h                     ; F9 key
                db      44h                     ; F10 key
                db      54h                     ; Shift F1 key
                db      55h                     ; Shift F2 key
                db      5Dh                     ; Shift F10 key
                db      85h                     ; F11 key
                db      86h                     ; F12 key
ENDIF

t1_11           db      00h                     ; ?

ITEMS_IN_TABLE1 equ     $ - table1              ; items in table for lookup

; Table 2 indexes correspond to table 1 indexes
; and points to code address where the function is handled.
table2:
                dw      offset error_return     ; ?

IFDEF FNKEY
                dw      offset fk12             ; Function Key 12
                dw      offset sfk10            ; Function Key 11
                dw      offset sfk10            ; Shift Function Key 10
                dw      offset sfk2             ; Shift Function Key 02
                dw      offset sfk1             ; Shift Function Key 01
                dw      offset fk10             ; Function Key 10
                dw      offset fk9              ; Function Key 09
                dw      offset fk8              ; Function Key 08
                dw      offset fk7              ; Function Key 07
                dw      offset fk6              ; Function Key 06
                dw      offset fk5              ; Function Key 05
                dw      offset fk4              ; Function Key 04
                dw      offset fk3              ; Function Key 03
                dw      offset fk2              ; Function Key 02
                dw      offset fk1              ; Function Key 01
ENDIF

IFDEF LINECAP
                dw      offset cmd_line_down    ; PgDn
                dw      offset cmd_line_up      ; PgUp
ENDIF

                dw      offset cntrl_page_up    ; Cntrl PgUp
                dw      offset cntrl_page_down  ; Cntrl PgDn
                dw      offset cntrl_home       ; Cntrl Home
                dw      offset cntrl_end        ; Cntrl End
                dw      offset cntrl_right_arrow; Cntrl Right Arrow
                dw      offset cntrl_left_arrow ; Cntrl Left Arrow
                dw      offset del_key          ; Del key
                dw      offset down_arrow       ; Down Arrow
                dw      offset end_key          ; End key
                dw      offset right_arrow      ; Right arrow
                dw      offset left_arrow       ; Left arrow
                dw      offset up_arrow         ; Up arrow
                dw      offset home_key         ; Home
                dw      offset backtab          ; Backtab

IFDEF FNKEY
fk6_data        db      'EXIT',CR
fk7_data        db      'DIR/P '
fk8_data        db      'COPY/V '
fk9_data        db      'C:\DOS\MORE< '
sfk1_data       db      'AUTOEXEC.BAT '
sfk2_data       db      'CONFIG.SYS '
sfk10_data      db      'C:\DOS\DISKCOPY '
fk11_data       db      'DISKCOPY '
fk12_data       db      'E:\DISKPARK',CR
ENDIF

new21int        equ     $
                CMP     AH,0AH                  ; Is this function 10?
                JNE     chain_on                ; nope, chain on to old int 21

                push    ax                      ; save ax
                push    bx                      ; save bx
IFNDEF AUX
                mov     bx, dx
                cmp     byte ptr ds:[bx], DOS_BUF_CHARS + 1  ; include CR here!
                jne     pop_out
ENDIF
                mov     ah, GET_PSP_ADDR        ; get psp segment of current program
                int     21h                     ; put it in bx
                push    es                      ; save es
                mov     es, bx                  ; move seg of psp to es
IFDEF LOGFILE
                mov     cur_psp, bx
ENDIF
                mov     bx, es:file_tbl_vec.l   ; get offset of file table
                mov     es, es:file_tbl_vec.h   ; get seg of file table
                mov     ax, es:[bx]             ; move first 2 handle entries to ax (stdin, stdout)
                pop     es                      ; restore es
                cmp     ax, 0101h               ; are stdin and stdout directed to CON?
                jz      stdin_out_ok            ; yes, do the one thing!
pop_out:
                pop     bx                      ; restore bx
                pop     ax                      ; restore ax
chain_on:
                jmp     dword ptr [oldint21vec] ; chain on!
stdin_out_ok:
                pop     bx                      ; restore bx
                pop     ax                      ; restore ax

                PUSH    AX
                PUSH    BX
                PUSH    CX
                PUSH    DX
                PUSH    SI
                PUSH    DI
                PUSH    BP
                PUSH    DS
                PUSH    ES
                STI
                CALL    begin_dosed
                POP     ES
                POP     DS
                POP     BP
                POP     DI
                POP     SI
                POP     DX
                POP     CX
                POP     BX
                POP     AX
                MOV     AL,00H
                IRET

;***************************************************************************
; Display and output services
;***************************************************************************
screen_and_display:                             ; Screen out control characters of form ^x
                CMP     AL, BLANK               ; Compare AL to blank character
                JNB     display_one_char        ; If we have a printable character go to
                                                ; display routine
                CMP     AL, TAB                 ; Is this a forward tab?
                JE      display_one_char        ; If so, go display.
                PUSH    AX                      ; No, this is a "normal" control character
                MOV     AL,HAT                  ; Place a "^" character in AL
                CALL    display_one_char        ; Print the "^" out.
                POP     AX                      ; Now convert the Control character to its
                OR      AL, UPPER_CASE_MASK     ; upper case equivalent and print it.
                                                ; Result is "^E"
                                                ; construct for representing control
                                                ; characters.
display_one_char:
                CMP     AL, BLANK               ; Does AL contain a blank character
                JB      display_cntrl_chr       ; Jump if we have a control character

ready_display:
                CMP     AL, CNTL_BACKSPACE      ; Don't increment column counter if
                JE      output_to_display       ; this is a control Backspace character
                INC     BYTE PTR CS:[current_col]; Increment column position because
                                                ; we received something printable
output_to_display:
                PUSH    DX                      ; Save the DX register and set up the
                XCHG    AX,DX                   ; AH register for a DOS function 2 (Display
                MOV     AH,  OUTPUT_CHR         ; output) call.
                INT     21H
                XCHG    AX,DX                   ; Restore the AX register
                POP     DX                      ; and restore the DX register
                RET

display_cntrl_chr       equ $                   ; Process a control character for output
                CMP     AL, CR                  ; Is this a carriage return character?
                JE      cr_out                  ; If so go take care of it
IFDEF NOTUSED
                                                ; see note by bs_out. TJR

                CMP     AL, BACKSPACE           ; Is this a backspace?
                JE      bs_out                  ; If so we need to take care of it
ENDIF
                CMP     AL, TAB                 ; Maybe its a tab character
                JNE     relay_to_output         ; Continue if its not a tab
                MOV     AL,BYTE PTR CS:[current_col] ; AL to current column on screen
                OR      AL,0F8H                 ; Get the number of spaces to the
                NEG     AL                      ; next 8 byte boundary
                PUSH    CX                      ; Save the CX register
                MOV     CL,AL                   ; Place the count of blanks to be output in CL
                moV     ch, 00h                 ; Zero out high byte
                JCXZ    skip_tabbing            ; Don't do anything if count is zero

tab_loop:
                MOV     AL, BLANK               ; Ok, now print out all necessary
                CALL    display_one_char        ; blanks to reach an 8 character
                LOOP    tab_loop                ; boundary.

skip_tabbing:
                POP     CX                      ; Restore the CX register.
                RET

cr_out:                                         ; Reset current column pointer to zero
                MOV     CS:BYTE PTR [current_col], 00H

relay_to_output:
                JMP     SHORT output_to_display

                                                    ; this code does seem to be accessed! TJR
IFDEF NOTUSED
    bs_out:
                    DEC     BYTE PTR CS:[current_col]   ; Decrement current column pointer
                    PUSH    AX                      ; Save the character in register AX
    IFNDEF NDEBUG
                    call    display_ax
    ENDIF
                    PUSH    DX                      ; Save the DX register
                    MOV     DX, ES:[0050H]          ; Get the Cursor position from Bios data area
                    MOV     AH, ES:[004AH]          ; Get # of cols on screen from Bios data area
    IFNDEF NDEBUG
                    call    display_ax
    ENDIF
                    OR      DL, DL                  ; Are we at column zero?
                    JNE     bs_end                  ; If not forget the rest.
                    OR      DH,DH                   ; Are off the screen at the top?
                    JNE     set_cur_row             ; If not then dont reset rows
                    ;MOV     DH, ROWS_ON_CRT         ; (Re)set row value to max
                    MOV     DH, es:byte ptr [crt_rows]  ; (Re)set row value to max
                    INC     DH                      ; inc so we can dec

    set_cur_row:
                    DEC     DH                      ; Decrement the row count
                    MOV     DL,AH                   ; Move in # of columns.  (0, 0) is home.
                    PUSH    BX                      ; Save BX
                    MOV     BH,00H                  ; Prepare for BIOS interrupt 10 funtion 2
                    MOV     AH, SET_CURSOR_POSN     ;    call (Set cursor position)
                    INT     10H                     ; Do it.
                    POP     BX                      ; Restore BX

    bs_end:
                    POP     DX                      ; Restore DX register
                    POP     AX                      ; and the AX register
                    JMP     SHORT relay_to_output   ; and now we are done.
ENDIF

decrement_col_row:                              ; Wrap back if line spans two screen lines
                OR      DL,DL                   ; Are we at column zero
                JNE     dcr_end                 ; If so then dont worry about row update
                DEC     DH                      ; Yep, decrement the row pointer
                MOV     DL,BYTE PTR ES:[crt_cols]  ; Get # of cols per screen from Bios data area

dcr_end:
                DEC     DL                      ; Decrement the column pointer
                RET

display_blank:
                MOV     AL, BLANK

display_and_test:                               ; Display and test for updated cursor location
                PUSH    CX                      ; Save CX register
                MOV     CX,ES:[cursor_posn]     ; Load CX with cursor position from Bios
                CALL    screen_and_display      ; Send the character in AL to output display
                CMP     CH, es:byte ptr [crt_rows]  ; Are we on the last row?
                ;CMP     CH, LAST_ROW_ON_CRT     ; Are we on the last row?
                JNE     dat_end                 ; If not then no further tests are needed
                CMP     CL,BYTE PTR ES:[cursor_posn]    ; Are we now on a new column?  (alternative question
                                                ; is: have we scrolled?)
                JBE     dat_end                 ; If current location > cl then yes we have scrolled
                                                ; up a line and need to decrement the line pointer
                DEC     DH                      ; Decrement the line pointer
IFDEF LINECAP
                test    bp,0010h                ; are we in line capture mode?
                jz      dat_end                 ; no, so don't worry
                mov     cx, cs:cur_cursor_pos   ; load old cursor position
                dec     ch                      ; decrement
                mov     cs:cur_cursor_pos, cx   ; store it again
ENDIF
dat_end:
                POP     CX                      ; Restore CX
                RET
;
;***************************************************************************
set_cursor:
                PUSH    BX
                PUSH    CX
                PUSH    DX
                PUSH    BP                      ; Save registers
                PUSH    SI
                PUSH    DI

inline_mod2:
                MOV     CX, MONO_CUR_DEF        ; Assume Standard char def
                                                ; (Note that this word is modified
                                                ; in line if it is later determined
                                                ; that a graphics board is in use)
                TEST    BYTE PTR ES:[kbd_flag], INSERT_ON_MASK ; Is insert on?
                JE      after_box               ; If not then branch
                MOV     CH,CL                   ; Yep, turn the cursor into
                SHR     CH,1                    ;   a half sized box

after_box:
                MOV     AH, SET_CURSOR_TYPE     ; Set up and call Bios for
                INT     10H                     ;   support.
                POP     DI
                POP     SI
                POP     BP                      ; Restore Registers
                POP     DX
                POP     CX
                POP     BX
                MOV     AX,WORD PTR ES:[cursor_posn]    ; Load AX with current cursor position

inline_mod3:
                CMP     AX,0888H                ; Compare current cursor position to end of line position
                JB      sc_ret                  ; If less than then dont update end of line position
                MOV     CS:WORD PTR [inline_mod3+1],AX  ; Modify end of line address with value of current cursor position

sc_ret:
                RET
;
begin_dosed:                                    ; Save Buffer Pointer register
                PUSH    DX
                MOV     BH,00H                  ; Set up to read current cursor position
                MOV     AH, READ_CURSOR_POSN    ; Cursor position returned as DH, DL is row, col
                INT     10H                     ; Read the current cursor position into DH, DL
                MOV     CS:[start_off], DL      ; Move current column into storage
                MOV     CS:[current_col], DL    ; Move current column into storage
IFDEF LINECAP
                mov     cs:cur_cursor_pos, dx   ; save cursor position
ENDIF
;;;             CMP     CL, MONO_CUR_STOP       ; Check: Is current end line for cursor = 12
;;;             JB      bd2                     ; Branch if less than 12
                MOV     AH, READ_VIDEO_STATE    ; Set up to return current video state
                INT     10H                     ; Go get it
                CMP     AL, GRAPHICS_CUR_STOP   ; Is this a 80x25 BW card
                MOV     CX, MONO_CUR_DEF        ; Set start/stop lines for cursor to 11/12
                JE      bd1                     ; Yes it is 80x25 BW card, branch
                cmp     ch,07                   ; Make sure cursor is one line only
                jz      bd2                     ; It already is
                MOV     CX, GRAPHICS_CUR_DEF    ; No its not 80x25 card, start/stop lines are 7/7
bd1:
                MOV     AH, SET_CURSOR_TYPE     ; Set function call to set cursor type
                INT     10H                     ; Do it

bd2:
                MOV     CS:WORD PTR [inline_mod2+1],CX  ; Save the cursor type setting in line
                POP     DX
                MOV     BX,DX                   ; DX points to offset of input buffer
                MOV     AL,[BX]                 ; Get the number of characters in the buffer
                CMP     AL,02H                  ; Is it bigger than 2 characters?
                JB      sc_ret                  ; Nope, we have a problem here!
                DEC     AL                      ; Now decrement AL to get number of real characters
                MOV     AH, 00H                 ;    buffer han hold.   Zero out AH
;IFNDEF NDEBUG
;                call    display_ax
;ENDIF
                MOV     CS:WORD PTR [inline_mod5+2],AX  ; AX now contains the max number of chars that the buffer
                                                ;    can hold.
                ADD     BX,02H                  ; BX points to address at which to put first real character
                                                ;    in input buffer

IFDEF AUXBUF
                                                ; AUTOMATIC BUFFER TOGGLE for COMMAND.COM access
auto_buffer_toggle:
                MOV     BP,DS                   ; Place the DS register into BP
                PUSH    DS                      ; Save the value of the DS register
                MOV     DI,CS                   ; Place the CS register into DI
                MOV     DS,DI                   ; and also into DS

;               CMP     BP,DI                   ; Compare DS : CS values
;               JB      abt1                    ; If DS < CS then jump
                                                ; this test does not work for new ver command.com
                                                ; as input buffer is moved to highRAM (9000-A000) area

                CMP     AL, DOS_BUF_CHARS       ; so now we check for input buffer size!
                JNE     abt1                    ; DOS does 7F bytes, if not 7F then jump

inline_mod4:
                CMP     DI,0888H                ; Compare CS : AUX
                                                ; (this in-line address will be called AUXiliary)
                JE      abt2                    ; If its equal (if we came this way last time!) then go swap
                JMP     SHORT after_swap_loop   ; If so then avoid swaps

abt1:
                MOV     BP,DI                   ; Set BP to highest register value
                                                ; (now determined to be CS)
                CMP     DI,DS:WORD PTR [inline_mod4+2]  ;  Compare CS : AUX
                JE      after_swap_loop         ; If its not equal (if we came this way
                                                ; last time!) then go swap

abt2:
                PUSH    CS                      ; Now we must swap the appropriate memory
                                                ; locations to allow the editor to use the
                POP     ES                      ; other buffer (There are two buffers that
                                                ; one can transfer between depending on
                                                ; DS : CS relationship.
                MOV     SI,offset last_entry    ; Load SI with starting address
                LEA     DI,[SI+08H]             ; Load DI with second starting address
                MOV     CX,0004H                ; 4 words are to be moved ... in order ...
                                                ; address of last string accessed,
                                                ; address of highest available byte, base
                                                ; address of text holding buffer,
                                                ; high address of holding buffer.
                                                ; Values are initiallized to:
                                                ; 1st buffer - 60h, 60h, 60h, 160h
                                                ; 2nd buffer - 160h, 160h, 160h, 260h
swap_loop:
                MOV     AX,[DI]                 ; Save the value in the secondary buffer
                MOVSW                           ; Move the primary buffer value into the
                                                ; secondary buffer position
                MOV     [SI-02H],AX             ; Replace emptied location with saved value
                LOOP    swap_loop               ; Go back and swap next set of values

after_swap_loop:
                MOV     DS:WORD PTR [inline_mod4+2],BP  ; Load AUX such that AUX = Max(CS, DS)
                POP     DS                      ;  Restore DS


                MOV     BP,000CH
                JMP     SHORT rib1
ENDIF

reinit_buffer:
                MOV     BP,0004H

rib1:
                XOR     SI,SI                   ; Zero out Source index
                MOV     CS:WORD PTR [inline_mod3+1],SI  ; Modify inline address just above
                XOR     DI,DI                   ; Zero out Destination index
                MOV     ES,DI                   ; Zero out the ES register
                AND     BYTE PTR ES:[kbd_flag], INSERT_OFF_MASK ; Turn insert off

start_read_char:
                AND     BP,0FFFEH               ; Turn off low bit in BP (Line may be shorter)
                CALL    set_cursor              ; Set the cursor shape as per mode
                XCHG    AX,DX                   ; Save AX in DX
                MOV     AH, INPUT_CHR_NOECHO    ; Get a character from the keyboard
                INT     21H                     ; Call DOS for service
                CMP     AL, ACK                 ; Is it an acknowledge character?
                JE      start_read_char         ; If so ... just forget it and go back
                MOV     CX, offset start_read_char  ; Save address of start_read_char label
                PUSH    CX                      ; This is a return address for a nonexistent call!
                OR      AL,AL                   ; Set up for test
                JNE     normal_keyboard_chr     ; Jump if non null character input

extended_keyboard_chr:
                MOV     AH, INPUT_CHR_NOECHO    ; Null character, get ready to read next character
                INT     21H                     ; Get it
                PUSH    ES                      ; Save ES
                PUSH    CS                      ; Set ES to value in CS
                POP     ES
                PUSH    DI                      ; Save DI
                MOV     CX, ITEMS_IN_TABLE1     ; Set counter to number of items in table 1
                MOV     DI, offset [table1]     ; Set starting address of table 1
                repnz   SCASB                   ; Scan for occurrence of second character in table 1
                POP     DI                      ; Restore DI
                POP     ES                      ;    and ES
                XCHG    CX,BX                   ; Let BX contain pointer to item found in table 1
                ADD     BX,BX                   ; Double it
                PUSH    CS:WORD PTR [BX+table2] ; Index into table 2 for starting address of service code
                                                ;    The value is pushed onto the stack to allow for the
                                                ;    return (Without a corresponding call!)
                MOV     BX,CX                   ; Restore original BX value
                RET                             ; Execute the fake return
;
;***************************************************************************
; Individual service routines
;***************************************************************************
;**** Escape Key
escape_handling:                                ; Escape character here
                CALL    home_key                ; Go to start of line
                CALL    cntrl_end               ; Delete to end

restart_input:
IFDEF LINECAP
                call    clear_line_mark
ENDIF
                POP     DI                      ; Pop the return address off of the stack
                JMP     SHORT reinit_buffer     ; Jump to reinitialize buffer and read

normal_keyboard_chr:
                CMP     AL, CNTL_BACKSPACE      ; Control Backspace?
                JE      nkc1                    ; if so, go to backspace handling
                CMP     AL, BACKSPACE           ; Backspace?
                JNE     nkc2                    ; If not continue

nkc1:
                JMP     backspace_handling      ; Backspace here, take care of it

nkc2:
                CMP     AL, CR                  ; Is it a carriage return?
                JNE     nkc3                    ; If not continue
                JMP     cr_key                  ; Yes, go take care of it

nkc3:
                CMP     AL, ESCAPE              ; Is it an escape character
                JE      escape_handling         ; If so, go take care of it

;;;             CMP     AL, TAB                 ; Bad code here                                              ** remove this **
;;;             JE      nkc4                    ; Bad code here                                              ** remove this **

inline_mod5:
nkc4:
                CMP     DI,0888H                ; Compare DI to max buffer size (modified in line in Begin_Dosed)
                JB      printable_char          ; If less than max then continue with printable character
                MOV     AL, BELL                ; Prepare to ring the bell
                JMP     display_one_char        ; Do it

;******************************************************************************
;***** Non editing character
printable_char:
                AND     BP,0FFF7H               ; Turn off the no changes made flag
                OR      BP,0002H                ; Turn on the changes made to this line flag
                PUSH    AX                      ; Save AX
                CALL    display_and_test        ; go display the character
                MOV     DX,ES:[cursor_posn]     ; Place the current cursor position in DX
                TEST    BYTE PTR ES:[kbd_flag], INSERT_ON_MASK  ; Is insert mode mask on?
                JE      add_chr_to_inbuf        ; If not then branch around shift up string
                PUSH    DI                      ; Save the DI register

shift_up:
                CMP     DI,SI                   ; Are we pointing past the end?
                JBE     end_shift_up            ; Have we decended to the original point?
                MOV     AL,[BX+DI-01H]          ; Move this character in input buffer up
                MOV     [BX+DI],AL              ;    one byte in memory
                DEC     DI                      ; Next character down
                JMP     SHORT shift_up          ; Return for more

end_shift_up:
                POP     DI                      ; Restore original end of input buffer pointer
                INC     DI                      ; Increment it by one
                INC     SI                      ; Also increment the current character pointer
                CALL    redisplay_line
                DEC     SI                      ; Now decrement source so that it will look
                                                ;    like normal overwrite to next code section

add_chr_to_inbuf:
                POP     AX                      ; Restore AX
                MOV     AH,AL                   ; Move the current character into AH
                XCHG    AL,[BX+SI]              ; Put it into the input buffer
                INC     SI                      ; Point to next character position
                CMP     DI,SI                   ; Have we gone past the end
                JNB     acti1                   ; If not then forget update of DI
                MOV     DI,SI                   ; DI is now reset to SI (which was >)

acti1:
                AND     AL,AH                   ; AND old character against new character
                CMP     AL, BLANK               ; If both are printable then we are finished
                                                ;    because we have coundn't have shortened
                JNB     acti_ret                ;    the line with the overwrite.  (eg. A over ^E
                                                ;    would shorten line by one character)
                TEST    BYTE PTR ES:[kbd_flag], INSERT_ON_MASK  ; If insert mode was on then we
                JNE     acti_ret                ;    are finished also because we have already
                                                ;    printed the string
                CMP     SI,DI                   ; Also we can forget if we are at the end of the line
                JE      acti_ret                ; None of the above hold, and we may have trailing
                JMP     note_shorter_change     ;    characters on the line that need to be blanked
                                                ;    over. (caused by overwriting a ^x diagraph or a
                                                ;    Tab character with a single character.)
acti_ret:
                RET

IFDEF FNKEY
;***************************************************************************
fk6             equ     $                       ; Function Key #06
                call    home_key
                call    cntrl_end
                cld                             ; execute the EXIT command
                mov     cs:fk_ptr,offset fk6_data
fk6_02          equ     $
                push    si
                mov     si,cs:fk_ptr
                lods    cs:byte ptr [si]
                mov     cs:fk_ptr,si
                pop     si
                cmp     al,CR
                jz      fk6_03
                call    nkc4
                jmp     fk6_02

fk6_03          equ     $
                jmp     cr_key

;***************************************************************************
fk12            equ     $                       ; Function Key #12
                call    home_key
                call    cntrl_end
                cld                             ; execute DISKPARK command
                mov     cs:fk_ptr,offset fk12_data
                jmp     fk6_02

;***************************************************************************
fk7             equ     $                       ; Function Key #07
                mov     cs:fk_ptr,offset fk7_data
fk7_02          equ     $
                cld
                call    home_key
          ;     call    cntrl_end
fk7_03          equ     $
                push    si
                mov     si,cs:fk_ptr
                lods    cs:byte ptr [si]
                mov     cs:fk_ptr,si
                pop     si
                cmp     al,BLANK
                jz      fk7_04
                call    nkc4
                jmp     fk7_03

fk7_04          equ     $
                jmp     nkc4

;***************************************************************************
fk8             equ     $                       ; Function Key #08
                mov     cs:fk_ptr,offset fk8_data
                jmp     fk7_02

;***************************************************************************
fk9             equ     $                       ; Function Key #09
                mov     cs:fk_ptr,offset fk9_data
                jmp     fk7_02

;***************************************************************************
sfk1            equ     $
                cld
                mov     cs:fk_ptr,offset sfk1_data
                jmp     fk7_03

;***************************************************************************
sfk2            equ     $
                cld
                mov     cs:fk_ptr,offset sfk2_data
                jmp     fk7_03

;***************************************************************************
sfk10           equ     $
                mov     cs:fk_ptr,offset sfk10_data
                jmp     fk7_02

;***************************************************************************
fk10            equ     $                       ; Function Key #10
                call    home_key
                call    cntrl_end
                mov     al,'C'
                call    nkc4
                mov     al,'L'
                call    nkc4
                mov     al,'S'
                call    nkc4
                mov     al,CR
                jmp     cr_key

;***************************************************************************
fk11            equ     $
                mov     cs:fk_ptr,offset fk11_data
                jmp     fk7_02

;***************************************************************************
fk1             equ     $                       ; Function Key #01
                mov     al,'A'
                call    nkc4
                mov     al,':'                  ; Display 'A:'
                jmp     nkc4

;***************************************************************************
fk2             equ     $                       ; Function Key #02
                mov     al,'B'
                call    nkc4
                mov     al,':'                  ; Display 'B:'
                jmp     nkc4

;***************************************************************************
fk3             equ     $                       ; Function Key #03
                mov     al,'C'
                call    nkc4
                mov     al,':'                  ; Display 'C:'
                jmp     nkc4

;***************************************************************************
fk4             equ     $                       ; Function Key #04
                mov     al,'D'
                call    nkc4
                mov     al,':'                  ; Display 'D:'
                jmp     nkc4

;***************************************************************************
fk5             equ     $                       ; Function Key #05
                mov     al,'E'
                call    nkc4
                mov     al,':'                  ; Display 'E:'
                jmp     nkc4

ENDIF

;***************************************************************************
;***** End key
end_key:
                TEST    BP,0008H                ; Have we made any changes to this buffer?
                JE      move_to_end             ; If so then go ahead an position to end of line.
                AND     BP,0FFF7H               ; No changes, lets now indicate that changes have been made,
                OR      DI,DI                   ;    and restore the input buffer.  Start by checking DI for zero.
                JNE     move_to_end             ; Go restore if something is there.
                MOV     AL,[BX-01H]             ; Get the length of the string in the holding buffer.
                CMP     [BX-02H],AL             ; Is it larger than the max buffer size?
                JB      recycle_input           ; If so then we have problems and a restart is necessary.
                MOV     AH,00H                  ; Zero out AH
                XCHG    AX,DI                   ; Move the new length into DI
                CMP     BYTE PTR [BX+DI], CR    ; Is there a carriage return here
                JE      move_to_end             ; If so then go ahead and restore it

recycle_input:
                JMP     restart_input           ; No, we must have garbage in the input buffer, go back and restart

move_to_end:
                CALL    right_arrow             ; move over one character to the right
                JB      move_to_end             ; keep going if not at end

error_return:
                RET

;***************************************************************************
;***** Control home key
cntrl_home:
                PUSH    SI                      ; Save the pointer into the input buffer
                CALL    home_key                ; Move to start of line
                POP     SI                      ; Restore pointer
                MOV     CX,DI                   ; Set CX to character count in string and remove the
                SUB     CX,SI                   ;    characters from start of line to current position from count.
                XOR     DI,DI                   ; Clear out the DI register
                PUSH    DI                      ; Save it.                                                               ** get rid of this **
                PUSH    CX                      ;   and the new count of characters.
                JCXZ    after_drop_down         ; If the new count of characters is zero then nothing else to do.

drop_text_down:
                MOV     AL,[BX+SI]              ; Get the current character
                INC     SI                      ; Advance source pointer to next character
                MOV     [BX+DI],AL              ; Store it down into the next destination available
                INC     DI                      ; Increment the destination pointer
                LOOP    drop_text_down          ;    and continue for the count

after_drop_down:
                POP     DI                      ; Restore DI
                POP     SI                      ;    and SI                                                             ** change this to xor SI,SI **
                JMP     note_shorter_change     ; Now go redisplay line on screen

;***************************************************************************
;***** Control end key
cntrl_end:
                OR      BP,0002H                ; Turn on the change to line flag
                MOV     CX,CS:WORD PTR [inline_mod3+1]  ; Get the screen position of the last character in the screen

blank_to_end:
                CMP     CX,ES:[cursor_posn]     ; Is the current screen position at the end of line?
                JBE     after_blanking          ; If so then jump
                CALL    display_blank           ; Display a blank
                JMP     SHORT blank_to_end      ; and continue until end of line position on screen is reached

after_blanking:
                MOV     DI,SI                   ; The new line length is current offset in the line
                JMP     place_cursor            ; and now place the cursor on the screen


;***************************************************************************
;***** Control right arrow
cntrl_right_arrow:                              ; Move right to start of next word
                CALL    right_arrow             ; Move right one character
                JNB     cra_ret                 ; If at the end then fall out of loop
                CALL    check_non_by_alpha      ; Check for non-alpha/alpha combination (return issued from subroutine
                                                ;    if the condition is satisfied)
                JMP     SHORT cntrl_right_arrow ; continue the cycle

cra_ret:
                RET

check_non_by_alpha:
                MOV     AL,[BX+SI]              ; Get character
                CALL    check_alpha             ; See if this is an alphabetic character
                JB      cnba_ret                ; If not then we are finished
                MOV     AL,[BX+SI-01H]          ; Get previous character
                CALL    check_alpha             ; Is this an alpha?
                JNB     cnba_ret                ; If so then we are still not happy and must continue searching.
                POP     AX                      ; Pop off the fake return address if this location points
                                                ;    to an alpha character and the previous one doesnt.
cnba_ret:
                RET

check_alpha:
                OR      AL, LOWER_CASE_MASK     ; Or in lower case bit
                CMP     AL, SMALLA              ; Is this the letter "a"?
                JB      cnba_ret                ; It is less than the letter "a" ... return
                CMP     AL, LEFT_CURLY_BRACE    ; Is this the character "{"?  (one greater than "z")
                CMC                             ; Complement Carry flag so its on if < "{"
                RET

;***************************************************************************
;***** Right arrow
right_arrow:
                CMP     SI,DI                   ; Are we past the last character
                JNE     redisplay_char          ; If not then continue
                RET

redisplay_char:
                MOV     AL,[BX+SI]              ; Move the current character into AL
                INC     SI                      ; Bump the pointer into input buffer
                CALL    display_and_test        ; Redisplay that character
                STC                             ; Set the carry flag
                RET

;***************************************************************************
;;;L0474        equ     $                       ;\                             ;
;;;             CALL    right_arrow             ; \                            ;
;;;             JNB     L0482                   ;  \                           ;
;;;             MOV     AL,07H                  ;   \ This code is not called. ;
;;;             AND     AL, BYTE PTR ES:[cursor_posn]                          ;
;;;             JNE     L0474                   ;  /                           ;
;;;L0482        equ     $                       ; /                            ;
;;;             RET                             ;/                             ;
;***************************************************************************

;***************************************************************************
;***** Control left arrow
cntrl_left_arrow:
                CALL    left_arrow              ; Move left one character
                JB      bt_ret                  ; Jump if we are at the end of the buffer
                CALL    check_non_by_alpha      ; See if we are at the start of an alph character string.
                JMP     SHORT cntrl_left_arrow  ; If not then loop (if we were then the return was issued from
                                                ;    check_non_by_alpha)

;***************************************************************************
;*** Backtab key
backtab:
                CALL    left_arrow              ; move left one character
                JB      bt_ret                  ; If at the start of the line then quit
                MOV     AL,07H                  ; set up the mask
                AND     AL,DL                   ; check to see if we are on an 8 byte column.
                JNE     backtab                 ; If not then continue left.
bt_ret:
                RET
;
;***************************************************************************
;***** Left Arrow
left_arrow:
                OR      SI,SI                   ; Are we at the start of the line
                STC                             ; Set the carry flag
                JE      cnba_ret                ; If so then do nothing
                DEC     SI                      ; Move the input buffer pointer back one
                CALL    decrement_col_row       ; Adjust the col row positon
                CMP     BYTE PTR [BX+SI], TAB   ; Have we just run over a tab?
                JE      back_over_tab           ; If so then go expand it
                CMP     BYTE PTR [BX+SI], BLANK ; Compare for printable character
                JNB     la_end                  ; If we have one then jump to final section
                CALL    decrement_col_row       ; Decrement col row one more time (for ^ character)
                JMP     SHORT la_end            ; Finish up
                                                ; Expand the tab character
back_over_tab:
                PUSH    SI                      ; Save the SI pointer into input buffer
                MOV     CX,SI                   ; Set the count to its value
                MOV     AL,CL                   ;    and also AL
                MOV     AH,07H                  ;
                JCXZ    no_prev_tabs            ; At start of line (no previous tab to align us)
                                                ; See if we can find another tab in front of us to
                                                ;    align us on screen (character after that tab
                                                ;    would be properly aligned on 8 byte boundary.)
tab_align:
                DEC     SI                      ; Decrement the pointer into input buffer
                CMP     BYTE PTR [BX+SI], BLANK ; Check for printable character
                JNB     tab_align_end           ; If found cycle in loop
                CMP     BYTE PTR [BX+SI], TAB   ; Have we succeded in finding another tab?
                JE      sub_prev_char_cnt       ; If so then we know we are aligned so jump
                DEC     AH                      ; No, this must be a control character diagraph (^x form)
                                                ;    so we need to subract an additional character from the
tab_align_end:                                  ;    count of characters found so far.
                LOOP    tab_align               ; Go back for more

no_prev_tabs:                                   ; Ok, we only get here if there were no previous
                                                ;    characters in input buffer or if there were no previous
                                                ;    tabs in the buffer
                SUB     AH,CS:[start_off]       ; Subtract out the screen bias position for location 0 of the
                                                ;    input buffer pointer.
sub_prev_char_cnt:
                SUB     AH,AL                   ; These two statements remove the count of all previous characters
                ADD     CL,AH                   ;    in the buffer (either to tab boundary or start of line).
                                                ;    Additional diagraph hats "^"
                                                ; have been previously subtracted from AH.
                AND     CL,07H                  ; take modulo 8
                JE      end_jump_tab            ; Did this tab expand into only one blank (remember we've already
                                                ;    moved over one!), if so jump
jump_tab:
                CALL    decrement_col_row       ; Calculate new column and row on screen
                LOOP    jump_tab

end_jump_tab:
                POP     SI                      ; Restore input buffer pointer

la_end:
                CLC                             ; Clear the carry flag
                JMP     SHORT place_cursor      ; Go place the cursor on the screen

;***************************************************************************
;***** Home key
home_key:                                       ; home key service begins here
                CALL    left_arrow              ; move over one
                JB      pc_ret                  ; jump if back at start
                JMP     SHORT home_key          ; Not there yet, keep going
;
;***************************************************************************
;***** Backspace key
backspace_handling:                             ; backspace handling begins here
                CALL    left_arrow              ; first just move over one character
                JB      pc_ret                  ; If we are already at the start of the line
                                                ;    then we neednt continue
;***************************************************************************
;***** Delete Key
del_key:                                        ; Delete key handling begins here
                OR      DI,DI                   ; Is the string length zero?
                JE      pc_ret                  ; If so go to end
                CMP     SI,DI                   ; Are we past the end of the string?
                JE      pc_ret                  ; If so go to end
                PUSH    SI                      ; Save the SI pointer to input buffer

shift_down:
                CMP     SI,DI                   ; Are we past the end:
                JE      after_shift_down        ; If so, finish up
                MOV     AL,[BX+SI+01H]          ; Get next character up
                MOV     [BX+SI],AL              ; then shift it down into this slot
                INC     SI                      ; Increment the input buffer pointer
                JMP     SHORT shift_down        ;    and cycle

after_shift_down:
                POP     SI                      ; Restore the input text buffer pointer
                DEC     DI                      ; Decrement the length of the input buffer string

note_shorter_change:
                OR      BP,0003H                ; Indicate that we have changed and potentially shortened the string.

redisplay_line:
                AND     BP,0FFF7H
                PUSH    SI                      ; Save the pointer into the input buffer

display_next_char:
                CMP     SI,DI                   ; Are we past the end of the string?
                JNB     after_display_char      ; If so dont display
                MOV     AL,[BX+SI]              ; Get the character from the input buffer
                CALL    display_and_test        ; send it to the display
                INC     SI                      ; update buffer pointer
                JMP     SHORT display_next_char ; continue to end

after_display_char:
                CALL    set_cursor              ; Set type of cursor, AX returns with current cursor location (end of line)
                TEST    BP,0001H                ; Is the old line longer?
                JE      after_blank_pad         ; If not then forget blanking the tail
                XCHG    AX,CS:WORD PTR [inline_mod3+1]  ; Put current cursor positon in end of line indicator and
                                                ;    save that value for blanking operation
                XCHG    AX,CX                   ; Put former end of line position in CX

add_blank_pad:
                CALL    display_blank           ; Display a blank
                CMP     CX,ES:[cursor_posn]     ; Is current cursor position now at end of line
                JNBE    add_blank_pad           ; If not, then continue blanking loop.

after_blank_pad:
                POP     SI

place_cursor:
                MOV     AH, SET_CURSOR_POSN     ; Load AH for Bios call
                PUSH    BX                      ; Save BX and CX registers
                PUSH    CX
                MOV     CS:[current_col],DL     ; Record current column position
                MOV     BH,00H
                INT     10H                     ; Call bios for service
                POP     CX                      ; Restore registers
                POP     BX

pc_ret:
                RET

;***************************************************************************
;***** Up Arrow
up_arrow:                                       ; Up arrow service starts here
                STD                             ; Set direction flag to decrement

;***************************************************************************
;***** Down Arrow
down_arrow:                                     ; Down arrow service starts here
IFDEF LINECAP
                call    clear_line_mark
ENDIF
                AND     BP,0FFF7H
                CALL    home_key                ; Go to beginning of line
                CALL    cntrl_end               ; Delete to end of line
;
da1:
                PUSH    ES                      ; Save ES register
                PUSH    DS                      ; Set ES to DS
                POP     ES
                PUSH    CS                      ; Set DS to CS
                POP     DS
                MOV     SI,DS:[last_entry]      ; Get pointer to first byte of current string in
                                                ;    input buffer (from text buffer)
                LODSB                           ; Obtain the character
                LEA     DI,[BX-01H]             ; Load DI with the offset of the length character
                                                ;    of the input buffer
                MOV     CX, BUF_SIZE         ; Set counter to look at each location in the text buffer

search_for_len_byte:
                CALL    get_char                ; Get a character from the text holding buffer
                TEST    AL,  LENGTH_BYTE_MASK   ; Is it a length byte?
                LOOPZ   search_for_len_byte     ; Nope, go back and try again
                CLD                             ; Yep, clear the direction flag
                JNE     found_a_len_byte        ; Jump if we found a length byte
                PUSH    ES                      ; Restore DS
                POP     DS
                POP     ES                      ; Restore ES
                JMP     restart_input

found_a_len_byte:
                AND     BP,0FFF1H               ; Indicate new line from holding buffer, and no changes made to it.
                CALL    get_char                ; Get a dummy character to readjust the SI pointer
                TEST    BYTE PTR [SI],  LENGTH_BYTE_MASK  ; Is next location a length byte
                JNE     recover_string          ; If so (then down arrow issued) jump to recover string
                DEC     SI                      ; No, an up arrow was issued and we need to back SI down two bytes to
                                                ;   set up for next get_char to be the length byte.
                CALL    swap_and_get_tp         ; Back a-one
                DEC     SI
                CALL    swap_and_get_tp         ; and a-two

recover_string:
                MOV     DS:[last_entry],SI      ; Load the address of the last accessed buffer string
                CALL    get_char                ; Get the length character from holding buffer
                AND     AL, LENGTH_OFF_MASK     ; Mask off the length byte identifier
                MOV     AH, BYTE PTR ES:[DI-01H]; Place Max buffer size in AH
                CMP     AL,AH                   ; Is text string smaller than max buffer size?
                JB      continue_recovery       ; If so then continue recovery
                MOV     AL,AH                   ; Uh-oh, take the max buffer size as the count of chars to restore

continue_recovery:
                STOSB                           ; Save the length count
                CBW                             ; Get rid of high portion of AX
                XCHG    AX,CX                   ; Place the count in CX
                PUSH    CX                      ;    and save it for later

load_chr_and_display:
                CALL    get_char                ; Get a character from the holding buffer
                STOSB                           ; Store this character in the input buffer
                CALL    display_and_test        ; Display this character
                LOOP    load_chr_and_display    ; Continue until all characters are done
                POP     DI                      ; Place the count of the number of characters in the
                                                ;    string into DI
                MOV     SI,DI                   ; Require the current pointer to be just past the end
                PUSH    ES                      ; Restore DS
                POP     DS
                POP     ES                      ; Restore ES

lcad_ret:
                RET

;***************************************************************************
;***** Control Page Down
cntrl_page_down:
                TEST    BP,0006H                ; Are we on the original line or have we made mods
                JNE     lcad_ret                ; If so then do nothing
                PUSH    DS                      ; Save some registers
                PUSH    ES
                PUSH    BX
                CALL    home_key                ; Move to start of line
                CALL    cntrl_end               ; Delete to the end of the line
                PUSH    CS                      ; Set ES register to CS                                                ** change to mov **
                POP     ES
                PUSH    CS                      ; Set DS register to CS
                POP     DS
                MOV     DI,DS:[last_entry]      ; Get address of last string accessed in holding buffer
                PUSH    DI                      ; Save it.
                MOV     SI,DI                   ; Point the source index to it.

skip:
                CALL    get_char                ; Get a character from the text buffer
                CMP     AL, CR                  ; Is it a carriage return?
                JNE     skip                    ; If not keep going

move_down:
                CMP     SI,DS:[next_entry]      ; Are we at the next available location in buffer
                JE      pad_crs                 ; If so then break out of loop
                CALL    get_char                ; Get this character
                CALL    save_char               ;    and move it down in the holding buffer.
                JMP     SHORT move_down         ; Continue the loop.

pad_crs:
                MOV     AL, CR                  ; Load AL with a carriage return
                MOV     [DI],AL                 ; and into the buffer
                MOV     BX,DI                   ; Now reset the base pointer one beyond last string in holding buffer.
                XCHG    BX,DS:[next_entry]      ; Put the old value in BX and the new value in place
                POP     SI                      ; Restore the pointer to last string accessed

continue_pad_cr:
                CMP     BX,DI                   ; Are we at the old next_available slot yet?
                JE      after_pad_cr            ; If yes the break out
                CALL    save_char               ; No, save the carriage return
                JMP     SHORT continue_pad_cr   ; Keep going

after_pad_cr:
                DEC     SI                      ; Point to just before this entry (so Down arrow will pick it up)
                MOV     DS:[last_entry],SI      ; Save it as last accessed text entry
                POP     BX                      ; Restore registers.
                POP     ES
                POP     DS
                JMP     da1                     ; Go perform a down arrow operation

;***************************************************************************
;***** Carriage Return
cr_key:
IFDEF LINECAP
                call    clear_line_mark
ENDIF
                AND     BYTE PTR ES:[kbd_flag],INSERT_OFF_MASK  ; Turn off insert mode
                CALL    set_cursor              ; Change the cursor on the screen
                add     sp,02                   ; Pop the fake return address off
                                                ; the stack (so we really return)
                CALL    move_to_end             ; Move the cursor to the end of the line
                MOV     AL,CR                   ; Move a CR into AL
                CALL    display_one_char        ; Display it
                MOV     [BX+DI],AL              ; Add the CR to the input buffer
                MOV     AX,DI                   ; Move the length into AX
                MOV     [BX-01H],AL             ; Now move it into the length byte of the input buffer
                MOV     SI,BX                   ; Point the Source index to the input buffer
IFDEF LOGFILE
                cmp     file_name_ptr, -1
                je      no_log_write
                push    ax
                push    bx
                push    cx
                push    dx
                push    ds

                mov     bx, cs
                mov     ah, SET_PSP_ADDR
                int     21h
IFNDEF NDEBUG
                xor     bx, bx
                mov     ah, GET_PSP_ADDR
                int     21h
ENDIF
                mov     ax, cs
                mov     ds, ax

                mov     dx, file_name_ptr       ; get file_name_ptr

                mov     al, 0C1h                ; bit 7=1 child does not inherit handle
                                                ; bits 4-6=100 deny none
                                                ; bit 3=0 reserved
                                                ; bits 0-2=001 write access
                mov     ah, DOS_OPEN_H
                int     21h
                jnc     log_file_opened

                xor     cx, cx                  ; bit 5=0 archive bit (wow!)
                mov     ah, DOS_CREATE_NEW_H
                int     21h
                jc      no_log_open
log_file_opened:
                mov     bx, ax                  ; mov handle into bx

                xor     cx, cx                  ; most sig offset
                xor     dx, dx                  ; least sig offset
                mov     al, 02h                 ; end of file relative
                mov     ah, DOS_SET_FILE_PTR    ; reset our file pointer
                int     21h

                call    dump_date

                mov     cx, di                  ; number of bytes to write
                inc     cx                      ; add the CR again
                xor     ch, ch                  ; not >= 128?
                mov     dx, si
                pop     ds                      ; restore ds
                push    ds                      ; from stack
                mov     ah, DOS_WRITE_H
                int     21h

                mov     cx, 01h                 ; one byte to write
                mov     dx, offset line_feed    ; output a LF to log
                mov     ax, cs
                mov     ds, ax
                mov     ah, DOS_WRITE_H
                int     21h

                mov     ah, DOS_CLOSE_H
                int     21h
no_log_open:
                mov     bx, cur_psp
                mov     ah, SET_PSP_ADDR
                int     21h

                pop     ds
                pop     dx
                pop     cx
                pop     bx
                pop     ax
no_log_write:
ENDIF
                CMP     AL,03h                  ; Are there at least two characters?   ** remove this? **
                JBE     crk_ret                 ; If not then just forget the save
IFDEF NOTSOME
                cmp     word ptr [bx],'ID'      ; Don't add the DIR/P command to the
                jne     not_dir_p               ; save buffer
                cmp     word ptr [bx+02],'/R'   ; Bytes are reversed in a word
                je      crk_ret

not_dir_p       equ     $
                cmp     word ptr [bx],'XE'      ; Don't add the EXIT command to the
                jne     not_exit                ; save buffer
                cmp     word ptr [bx+02],'TI'
                je      crk_ret
;
not_exit        equ     $
                cmp     word ptr [bx],'LC'      ; Don't add CLS either
                jne     not_cls
                cmp     byte ptr [bx+02],'S'
                je      crk_ret
;
not_cls         equ     $
ENDIF
                PUSH    CS                      ; Set ES to value in CS
                POP     ES
                INC     WORD PTR CS:[last_entry]; Bump the pointer to the strings location in the text buffer
                TEST    BP,0002H                ; Have changes been made to this line?
                JE      crk_ret                 ; If not then we are through
                MOV     DI,CS:[next_entry]      ; Point the destination to the next available slot
                OR      AL, LENGTH_BYTE_MASK    ; Add in the length byte bit mask

save_string:
                CALL    save_char               ; Add it to the holding buffer
                CMP     AL, CR                  ; Is this a carriage return?
                JE      after_save_string       ; If so, break out of loop
                LODSB                           ; Get the next character from the input buffer
                JMP     SHORT save_string

after_save_string:
                MOV     CS:[next_entry],DI      ; Save the new pointer to the text holding buffer

omit_partial_string:
                CMP     CS:[DI],AL              ; Is it a carriage return?
                JE      after_omit_partial      ; If so we are finished
                CALL    save_char               ; No, that means there is a partial string here
                JMP     SHORT omit_partial_string   ;   so fill with CRs until next CR is found

after_omit_partial:
                MOV     CS:[last_entry],DI

crk_ret:
                RET

get_char:                                       ; Obtain a character from the text holding buffer
                CALL    swap_and_get_tp         ; First we must reverse SI and DI pointers
                LODS    CS:BYTE PTR [SI]        ; Load the character from the text holding buffer into AL

swap_and_get_tp:                                ; Reverse SI and DI pointers because
                XCHG    SI,DI                   ;    get_text_pointer assumes that DI indexes
                CALL    get_text_pointer        ;    into the text holding buffer and not SI
                XCHG    SI,DI                   ; Verify legal pointer into holding buffer
                RET

save_char:                                      ; Add a character to the text holding buffer
                CALL    get_text_pointer        ; Verify correctness of pointer
                STOSB                           ; Move the character in AL to the text holding buffer

get_text_pointer:                               ; Obtain a legal pointer into the text holding buffer
                CMP     DI,CS:[buffer_bot]      ; Is the current pointer too small?
                JNB     gtp1                    ; If not, then continue
                MOV     DI,CS:[buffer_top]      ; Yes, set to upper limit
                DEC     DI                      ;    and decrement one

gtp1:
                CMP     DI,CS:[buffer_top]      ; Is the current pointer too big?
                JB      gtp_ret                 ; If not jump to end
                MOV     DI,CS:[buffer_bot]      ; Yes, set to lower limit

gtp_ret:
                RET
IFDEF LOGFILE
;***************************************************************************
; call with bx = file_handle
date_data:
dd_month        dw      '00'
                db      '/'
dd_day          dw      '00'
                db      '/'
dd_year1        dw      '00'
dd_year2        dw      '00'
                db      ' '
dd_hour         dw      '00'
                db      ':'
dd_minute       dw      '00'
                db      ':'
dd_second       dw      '00'
                db      '.'
dd_hund         dw      '00'
                db      ' '
date_data_size  equ     $ - offset date_data
dump_date:
                push    ax
                push    cx
                push    dx
                push    ds
                push    bx

                mov     ax, cs
                mov     ds, ax

                mov     ah, DOS_GET_DATE
                int     21h

                xor     ah, ah
                mov     al, dh
                call    ax_to_2char
                mov     dd_month, ax

                xor     ah, ah
                mov     al, dl
                call    ax_to_2char
                mov     dd_day, ax

                mov     ax, cx
                mov     cx, 100
                div     cl
                push    ax
                xor     ah, ah
                call    ax_to_2char
                mov     dd_year1, ax
                pop     ax
                xchg    ah, al
                xor     ah, ah
                call    ax_to_2char
                mov     dd_year2, ax

                mov     ah, DOS_GET_TIME
                int     21h

                xor     ah, ah
                mov     al, ch
                call    ax_to_2char
                mov     dd_hour, ax

                xor     ah, ah
                mov     al, cl
                call    ax_to_2char
                mov     dd_minute, ax

                xor     ah, ah
                mov     al, dh
                call    ax_to_2char
                mov     dd_second, ax

                xor     ah, ah
                mov     al, dl
                call    ax_to_2char
                mov     dd_hund, ax

                mov     cx, date_data_size      ; bytes to write
                mov     dx, offset date_data    ; output data/time to log
                mov     ah, DOS_WRITE_H
                int     21h

                pop     bx
                pop     ds
                pop     dx
                pop     cx
                pop     ax
                ret
ax_to_2char:
                push    cx
                mov     cl, 10
                div     cl                      ; quo - al, rem - ah
                or      ah, 30h
                or      al, 30h
                pop     cx
                ret
ENDIF
;***************************************************************************
;***** Control Page Up
cntrl_page_up:                                  ; Control Page Up service begins here
                CALL    home_key                ; Go to start of line
                CALL    cntrl_end               ; And delete forward
                PUSH    DS                      ; Save DS
                MOV     AX,CS                   ; Set DS and ES to value in CS
                MOV     DS,AX
                MOV     ES,AX
                MOV     DI,DS:[buffer_bot]      ; Get base address of text buffer
                MOV     DS:[last_entry],DI      ; Set current pointer into text buffer to base
                MOV     DS:[next_entry],DI      ; Set next new pointer into text buffer to base
                MOV     CX, BUF_SIZE         ; Set count to size of text holding buffer
                MOV     AL,CR                   ; Set AL with CR and fill holding
                repz    STOSB                   ; buffer with carriage returns
                POP     DS                      ; Restore DS
                JMP     restart_input

IFDEF LINECAP
;***************************************************************************
;***** Page Down
cmd_line_down:
                mov     ah,1                    ; move 1 line down
                jmp     short move_line_ptr     ; continue with line move
                nop
;***************************************************************************
;***** Page Up
cmd_line_up:
                mov     ah,0FFh                 ; move -1 lines (1 up)

move_line_ptr:
                push    ax                      ; save lines to move (ah)
                call    home_key                ; Go to start of line
                call    cntrl_end               ; and delete forward
                pop     ax                      ; restore lines to move (ah)
                test    bp,10h                  ; do we have a line mark now?
                jz      no_old_mark             ; no need to clear old one
                call    clear_line_mark
no_old_mark:
                push    bx
                push    dx
                push    es
                mov     al, dh
                mov     dx, cs:cur_cursor_pos   ; get our mark position
                xor     dl, dl                  ; clear the column portion
skip_cur_line:
                add     dh, ah                  ; inc,dec appropriately
                or      bp, 18h                 ; now we have a marker (10)
                                                ; and cursor has moved (08)
                cmp     dh, 0                   ; are we less than 0?
                jge     not_off_top             ; we didn't go off the top
                mov     dh, es:byte ptr [crt_rows]  ; set row to bottom line
                ;mov     dh, LAST_ROW_ON_CRT     ; set row to bottom line
not_off_top:
                cmp     dh, es:byte ptr [crt_rows]  ; > bottom line?
                ;cmp     dh, LAST_ROW_ON_CRT     ; > bottom line?
                jle     not_off_bottom          ; not off bottom of screen
                mov     dh, 0                   ; set row to top line
not_off_bottom:
                cmp     dh, al                  ; on editing line?
                jne     cur_line_ok             ; Jump if not equal
                jmp     short skip_cur_line
cur_line_ok:
                mov     cs:cur_cursor_pos, dx   ; save marker position
                mov     ax, es:crt_cols         ; get cols on screen
                cmp     al, [bx-02h]            ; is it smaller than size of buffer
                jb      buf_size_ok             ; yep, it's ok
                mov     al, [bx-02h]            ; nope, only copy (buffer size) bytes
                dec     al                      ; minus one.  we need space for CR
buf_size_ok:
                xor     ah,ah                   ; we can't do > 256 cols
                mov     si,ax                   ; set up for char copy
                push    si
                lea     di, [bx]                ; Load DI with addr of buffer
                mov     cs:word ptr [inline_mod6+1], si ; save num copied
                mov     bx, 70h
                mov     cx, 2000h
                mov     ah,1
                int     10h                     ; Video display   ah=functn 01h
                                                ;  set cursor mode in cx
                mov     cx,ds
                mov     es,cx
                mov     cx,1
loc_05A1:
                mov     ah,2
                int     10h                     ; Video display   ah=functn 02h
                                                ;  set cursor location in dx
                mov     ah,8
                int     10h                     ; Video display   ah=functn 08h
                                                ;  get char al & attrib ah @curs
                stosb                           ; Store al to es:[di]
                cmp     dl,0
                jne     loc_05B8                ; Jump if not equal
                mov     cs:byte ptr [inline_mod7+1], ah
                mov     ah,9
                int     10h                     ; Video display   ah=functn 09h
                                                ;  set char al & attrib bl @curs
loc_05B8:
                dec     si
                or      al,20h                  ; ' '
                cmp     al,20h                  ; ' '
                je      loc_05C4                ; Jump if equal
                mov     cs:word ptr [inline_mod6+1], si
loc_05C4:
                inc     dl
                cmp     si,0
                jne     loc_05A1                ; Jump if not equal
                pop     ax
inline_mod6:
                sub     ax, 0000h               ; subtract bytes copied
                mov     di, ax
                mov     cx, ax
                pop     es
                pop     dx
                mov     ah,2
                int     10h                     ; Video display   ah=functn 02h
                                                ;  set cursor location in dx
                pop     bx
                mov     [bx-1],cl
                mov     cx, word ptr cs:[inline_mod2+1]  ; Set the cursor type
                mov     ah,1
                int     10h                     ; Video display   ah=functn 01h
                                                ;  set cursor mode in cx
                jmp     move_to_end

;==========================================================================
;                              SUBROUTINE
;==========================================================================

clear_line_mark proc    near
                test    bp,10h
                jz      clm_ret                 ; Jump if zero
                and     bp,0FFEFh
                push    ax
                push    bx
                push    cx
                push    dx
                mov     dx,cs:cur_cursor_pos        ; (74E6:0160=0)
                xor     dl,dl                   ; Zero register
inline_mod7:
                mov     bx,7
                mov     cx,1
                mov     ah,2
                int     10h                     ; Video display   ah=functn 02h
                                                ;  set cursor location in dx
                mov     ah,8
                int     10h                     ; Video display   ah=functn 08h
                                                ;  get char al & attrib ah @curs
                mov     ah,9
                int     10h                     ; Video display   ah=functn 09h
                                                ;  set char al & attrib bl @curs
                pop     dx
                mov     ah,2
                int     10h                     ; Video display   ah=functn 02h
                                                ;  set cursor location in dx
                pop     cx
                pop     bx
                pop     ax
clm_ret:
                retn
clear_line_mark endp
ENDIF

IFNDEF NDEBUG
display_ax:
                push    dx
                push    ax
                xchg    dx,ax
                mov     ah, 02h
                mov     dl, dh
                shr     dl, 1
                shr     dl, 1
                shr     dl, 1
                shr     dl, 1
                or      dl, 30h
                cmp     dl, 39h
                jle     char0_ok
                add     dl, 07h
char0_ok:
                int     21h
                mov     dl, dh
                and     dl, 0fh
                or      dl, 30h
                cmp     dl, 39h
                jle     char1_ok
                add     dl, 07h
char1_ok:
                int     21h
                pop     dx
                push    dx
                shr     dl, 1
                shr     dl, 1
                shr     dl, 1
                shr     dl, 1
                or      dl, 30h
                cmp     dl, 39h
                jle     char2_ok
                add     dl, 07h
char2_ok:
                int     21h
                pop     dx
                push    dx
                and     dl, 0fh
                or      dl, 30h
                cmp     dl, 39h
                jle     char3_ok
                add     dl, 07h
char3_ok:
                int     21h
                pop     ax
                pop     dx
                ret
; test routine
buff            db      80h
                db      0
                db      80h dup (0)
test_prog:
                push    cs
                pop     ds
                mov     dx, offset buff
                mov     ah, 0ah
                int     21h
                int     20h
ENDIF

;***************************************************************************
inst_msg        db      'DOSEDIT (TJR v0.02) is now installed.',cr,lf
                db      '  options: '
IFDEF LOGFILE
                db      'LogFile '
ENDIF
IFDEF AUXBUF
                db      'AuxBuf '
ENDIF
IFDEF FNKEY
                db      'FnKey '
ENDIF
IFDEF LINECAP
                db      'LineCap '
ENDIF
IFDEF NOTSOME
                db      'NotSome '
ENDIF
IFNDEF NDEBUG
                db      'Debug'
ENDIF
                db      cr,lf,'$'
deinst_msg      db      'DOSEDIT deinstalled.',cr,lf,'$'

;
;***************************************************************************
deinstall       equ     $
                cld                             ; auto-increment index registers
                xor     ch, ch
                mov     cl, paramline           ; number of parameter characters in CX
                jcxz    deinstall02             ; zero parameter characters, just exit

                mov     bp, es                  ; segment of interrupt 21h vector
                mov     ax, ds                  ; call attention of ES to our neighborhood
                mov     es, ax

                mov     di, offset paramline+01
                mov     al, '/'                 ; search for one of these
                repnz   scasb
                jnz     deinstall02             ; slash not found, just exit

                mov     al, [di]                ; see about the next character
                and     al, 0DFh                ; make sure it is in upper case
                cmp     al, 'R'                 ; OK, is it the letter 'R'?
                jnz     deinstall02             ; no, again exit

                mov     es, bp                  ; ES equals installed DOSEDIT segment
                mov     ds, es:oldint21vec.h    ; get seg of old int 21 vector
                mov     dx, es:oldint21vec.l    ; get offset of old int 21 vector
                mov     ax, 2521h               ; restore original int 21h vector
                int     21h

                mov     ah, 49h                 ; deallocate installed DOSEDIT segment
                int     21h
                mov     ax, cs                  ; restore data segment
                mov     ds, ax

                mov     dx, offset deinst_msg
                mov     ah, 09
                int     21h

deinstall02     equ     $
                mov     ax, 4C00h
                int     21h                     ; Hey, it's Millertime
;***************************************************************************
install label   near
                mov     deflag, 'DE'            ; put this marker here before I forget

                mov     ax,3521h                ; get interrupt 21h vector in ES:BX
                int     21h

                cmp     es:deflag, 'DE'         ; OK, has DOSEDIT already been installed?
                jz      deinstall               ; yes, see about deinstalling it

                mov     oldint21vec.h, es       ; no, save original int 21h vector segment
                mov     oldint21vec.l, bx       ; and offset

                mov     ax, ds                  ; restore original ES
                mov     es, ax
IFDEF LOGFILE
                mov     file_name_ptr, -1
                cld                             ; auto-increment index registers
                xor     ch, ch
                mov     cl, paramline           ; number of parameter characters in CX
                jcxz    inst_no_param           ; zero parameter characters, just exit

                mov     di, offset paramline+01
                mov     al, '/'                 ; search for one of these
                repnz   scasb
                jnz     inst_no_param           ; slash not found, just exit

                mov     ax, [di]                ; see about the next characters
                and     al, 0DFh                ; make sure it is in upper case
                cmp     ax, ':F'                ; OK, is it 'F:'?
                jnz     inst_no_param           ; no, again exit
                add     di, 02h
                mov     dx, di                  ; file_name_ptr into dx
                mov     al, CR                  ; CR at end of file name
                repnz   scasb                   ; find it
                jnz     inst_no_param
                mov     byte ptr [di-1], 00h
                mov     file_name_ptr, dx       ; save file_name_ptr

                mov     al, 0C1h                ; bit 7=1 child does not inherit handle
                                                ; bits 4-6=100 deny none
                                                ; bit 3=0 reserved
                                                ; bits 0-2=001 write access
                mov     ah, DOS_OPEN_H
                int     21h
                jnc     file_opened

                xor     cx, cx                  ; bit 5=0 archive bit (wow!)
                mov     ah, DOS_CREATE_NEW_H
                int     21h
                jc      inst_no_param
file_opened:
                mov     bx, ax                  ; mov handle into bx
                mov     ah, DOS_CLOSE_H
                int     21h
inst_no_param:
ENDIF
                mov     dx, offset new21int
                mov     ax, 2521h               ; establish our own interrupt 21h vector
                int     21h

                                                ; initialize command buffer with CRs
                mov     di, offset puffer
                mov     cx, ALL_BUF_SIZE / 2 ; / 2 because we fill in twos!
                mov     ax, 0D0Dh               ; two carriage returns
                rep     stosw

                mov     es, envrnmt_seg
                mov     ah, 49h
                int     21h                     ; deallocate our environment segment

                mov     ah, 09                  ; print install message
                mov     dx, offset inst_msg
                int     21h
                                                ; get rid of this part of program
        ;;;     mov     dx, offset inst_msg     ; we just did this.
                int     27h                     ; terminate and stay resident
dosedit         endp
;***************************************************************************
cseg            ends
                end     dosedit
